  'MLX90621 IR camera support
  
  'Verion control
  ' V0.1            ' first version to work
  ' V0.2            'simple graphical representation on VGA screen
  '                 expected bug : left side pixels show low values, mid and right okay
  ' V0.3            'used for measurements
  
  'dow we need debug ?
  debug=0         'show intermediate results and cal values
  debug2=1        'show end result
  debug3=0        'show sensor data
  
  
  
  'array defines
  'note: versus Melexis documentation i and j are swapped in this program
  const SA=&H60       'I2C 7 bit slave address is &h60
  const EEA=&h50      'I2C EEPROM base address
  DIM da%(2*16*4-1)   '2 bytes for each cell in the 16x4 array
  dim cel%(15,3)      'the individual photocells IR value
  dim VIR!(15,3)      'calculation array with floats
  dim cal%(255)       'to store calibration numbers read from EEPROM
  dim Bi!(15,3)       'fast lookup table
  Dim Ai!(15,3)       'fast lookup table
  dim ALi!(15,3)      'fast lookup table
  dim Tob!(15,3)      'actual object temperature matrix
  cols%=6             'number of colors
  dim colr%(cols%-1) = (rgb(blue),rgb(cerulean),rgb(red),rgb(orange),rgb(yellow),rgb(white)) 'tile colors
  dim colf%(cols%-1) = (rgb(white),rgb(white),rgb(white),rgb(blue),rgb(blue),rgb(black)) 'ink colors
  dim colm%(15,3)     'colors belonging to temperatures
  
  
  'text colors
  wh=rgb(white)
  gr=rgb(green)
  cy=rgb(cyan)
  
  
  
  'System initialization ---------------------------------------------
  
  'I2C bus open at 100kHz (max for MLX90614)
  setpin gp0,gp1,I2C
  I2C open 100,100
  
  
  'read calibration values from EEPROM
  read_cal
  process_cal
  if debug then show_cal
  
  
  'Initialize camera chip
  write_osc
  write_conf
  pause 100
  
  
  'user interface -----------------------------------------------------------
  cls
  text 0,0,"Peak logging tool for MLX90621 sensor",,,,gr
  text 0,24,"Tambient          Tsensor",,,,wh
  text 0,60,"Type any key to start logging",,,,wh
  
  do
    do_Ta
    dostuff
    tiles
    text 0,36,str$(Ta,4,2)+"          "+str$(calc_max!(),4,2),,,,cy
    pause 650
  loop while inkey$=""
  
  
  'define log file
  text 0,84,"Type LOG file name (no suffix), then <enter> "
  input a$
  text 0,96,"Storing data in A:/"+a$+".csv"
  
  
  'open log file and print header
  open "a:/"+a$+".csv" for output as #1
  print #1,"log data from MLX90614-DCI sensor"
  print #1,date$;",";time$
  
  
  'define log time
  Tick=1000 'ms
  Trun=60000 'ms
  
  
  'This is the timing definition of the test
  Tnew=timer+Tick
  Tend=Timer+Trun
    
  
  
  'main loop -----------------------------------------------------
  
  i=0
  do
    'Read ambient temperature from chip
    do_Ta
    
    'get temp data
    dostuff
    tiles
    
    cmax=calc_max!()
    cmin=calc_min!()
    cmean=calc_mean!()
    
    wrt$=str$(cmax,4,2)+","+str$(cmean,4,2)+","+str$(cmin,4,2)
    print #1,wrt$
    
    text 0,120,"Actual Max  = "+str$(calc_max!(),4,2)+" C",,,,cy
    text 0,144,"Actual Min  = "+str$(calc_min!(),4,2)+" C",,,,cy
    text 0,132,"Actual Mean = "+str$(calc_mean!(),4,2)+" C",,,,cy
 
    inc i
    
    do : loop while timer<Tnew     'wait for time to pass
    Tnew=timer+Tick
    
  loop while Timer<Tend            'until end of run
  
  close #1
  text 0,120,"                              "
  text 0,132,"          Done                ",,,,gr
  text 0,144,"                              "
end
  
  
  
  'subs and functions ---------------------------------------------
function calc_max!()
  calc_max=math(max Tob!())
end function

function calc_min!()
  calc_min=math(min Tob!())
end function

function calc_mean!()
  calc_mean=math(mean Tob!())
end function
  
  
sub do_Ta
  ptat_read
  calc_ta
end sub
  
  
sub dostuff
  'Read compensation pixel
  compix_read
  fast_lookup 'get constants in their final format
  
  'get data from camera
  get_IR
  convert
end sub
  
  
  'read MLX90621 IR sensors
sub get_IR
  
  'write: command,start addr,increment,number of reads = 16x4
  cmmd%=2
  I2C write SA,1,4,cmmd%,0,1,64
  
  'read the 64 16-bit values in da%() and fill the pixel array cel%()
  I2C read SA,0,128,da%()
  
  for i=0 to 15
    for j=0 to 3
      cel%(i,j) = da%(2*j+8*i) + da%(2*j+1+8*i)*256
      if cel%(i,j)>32768 then inc cel%(i,j),-65536            '2th complement
      if debug3 then print right$("000"+hex$(cel%(i,j) and &hffff),4);" ";
    next
    if debug3 then print
  next
  if debug3 then print
  
end sub
  
  'convert the IR values to temperatures in degrees C
sub convert
  print @(0,240);
  for j=0 to 3
    for i=0 to 15
      
      'if i=8 and j=2 and debug2=1 then
      'print:print Ai!(i,j),Bi!(i,j),Acom:print
      'end if
      
      VIR!(i,j)=cel%(i,j)-(Ai!(i,j) + Bi!(i,j)*(Ta-25)) 'Offset compensated
      
      'when TGC=0 and Eps=1 then simplify
      'VIR!(i,j)=(VIR!(i,j)-TGC_l!*Vcp_l!) / Eps         'Offset+TGC compensated
      
      'for BAB sensor Ks4_l!=0 -> Sx=0 -> simplify calculation
      'Sx=Ks4_l! * sqr(sqr((ALi!(i,j)^3 * VIR!(i,j)) + (ALi!(i,j)^4 * TaK4_l!)))
      'Tob!(i,j)=sqr(sqr((VIR!(i,j)/(ALi!(i,j)*(1 - 273.15*Ks4_l!)+Sx))+TaK4_l!))-273.15
      
      Tob!(i,j) = (VIR!(i,j)/ALi!(i,j) + TaK4_l!)^(0.25) - 273.15
      
      if debug2 then print str$(Tob!(i,j),3,1);
    next
    if debug2 then print:print:print
  next
  'if debug2 then print
  
end sub
  
sub tiles
  
  'font = 8x12 pixels
  'screen width = 640, 16 pixels => 40 pixels wide=5*8
  'height tile = 3x12 pixels
  
  'map values to color matrix
  math window Tob!(),0,cols%-1,colm%()
  
  
  'show colors
  for j=0 to 3
    for i=0 to 15
      tile i*5,19+j*3,colf%(colm%(i,j) ),colr%(colm%(i,j)),5,3
    next
  next
end sub
  
  'create fast lookup tables
sub fast_lookup
  
  KsTa_l! = KsTa / 2^20             '= 0 ????
  TGC_l! = TGC / 32                 '= 0 ????
  Ks4_l! = Ks4 / Ks4_sc             '= 0 for -BAB and -BAD sensors in datasheet
  Vcp_l! = Vcp - (Acp + Bcp*(Ta-25))
  ALcp_l! = ALcp / AL0_sc
  TaK4_l! = (Ta + 273.15)^4
  
  if debug then
    print "KsTa_l = ";KsTa_l!
    print "TGC_l  = ";TGC__l!
    print "Ks4_l  = ";Ks4_l!
    print "Vcp_l  = ";Vcp_l!
    print "ALcp_l = ";ALcp_l!
    print "TaK4   = ";TaK4_l!
  end if
  
  for i=0 to 15
    for j=0 to 3
      'Bi
      Bi!(i,j)=cal%(64+j+4*i) : if Bi!(i,j)>127 then inc Bi!(i,j),-256
      Bi!(i,j)=Bi!(i,j)/Bi_sc
      
      'Ai
      Ai!(i,j) =Acom + cal%(j+4*i)*dAi_sc
      
      'ALi (alpha)
      ALi!(i,j)=cal%(128+j+4*i)/dAL_sc + AL0/AL0_sc
      'when TGC=0 and KsTa=0 then skip next compensation
      'ALi!(i,j)=(ALi!(i,j)-TGC_l!*ALcp_l!)*(1 + KsTa_l!*(Ta-25))
      
    next
  next
  
end sub
  
sub read_cal
  'write eeprom start address for read = 0
  I2C write EEA,0,1,0
  
  'read the 256 bytes
  I2C read EEA,1,255,cal%()   '255
  I2C read EEA,0,1,cal%(255)  'the last one(256)
  'if debug then math v_print cal%()
  
end sub
  
sub show_cal
  ad=0
  print "d_Ai"
  for i=0 to 3
    for j=0 to 15
      print right$("  "+str$(cal%(ad+i+4*j)),3);" ";
    next
    print
  next
  print : ad=&h40
  print "Bi"
  for i=0 to 3
    for j=0 to 15
      print right$("  "+str$(cal%(ad+i+4*j)),3);" ";
    next
    print
  next
  print : ad=&h80
  print "d_alpha"
  for i=0 to 3
    for j=0 to 15
      print right$("  "+str$(cal%(ad+i+4*j)),3);" ";
    next
    print
  next
  print
  print "OSC_Trim = &h";hex$(osctrim)
  print "Conf Reg = &h";hex$(confreg)
  print "KsTa     = ";KsTa
  print "Ks4      = ";Ks4
  print "Ks4_sc   = ";Ks4_sc
  print "A_common = ";Acom
  print "Kt Scale = &h";hex$(KTscale)
  print "TGC      = ";TGC
  print "Emissivit= ";Eps
  print "dAi_sc   = ";dAi_sc
  print "Bi_sc    = ";Bi_sc
  print "Acp      = ";Acp
  print "Bcp      = ";Bcp
  print "ALcp     = ";ALcp
  print "Vth      = ";Vth
  print "Kt1      = ";Kt1
  print "Kt2      = ";Kt2
  print "dAL_sc   = ";dAL_sc
  print "AL0      = ";AL0
  print "AL0_sc   = ";AL0_sc
  
  print
  
end sub
  
  'calculate the proces constants from the data in the CAL array in EEPROM
sub process_cal
  
  'determine calculation constants
  osctrim=cal%(&hf7)
  confreg=cal%(&hf5)+cal%(&hf6)*256
  conf_L=cal%(&hf5):conf_H=cal%(&hf6)
  KsTa=cal%(&he6)+256*cal%(&he7):if KsTa>32767 then inc KsTa,-65536'2th complement
  Ks4_sc=2^((cal%(&hc0) and 15)+8)
  Ks4=cal%(&hc4):if Ks4>127 then inc Ks4,-256'2th complement
  Acom=cal%(&hd0)+cal%(&hd1)*256:if Acom>32767 then inc Acom,-65536'2th complement
  KTscale=cal%(&hd2)
  Acp=cal%(&hd3)+256*cal%(&hd4):if Acp>32767 then inc Acp,-65536'2th complement
  Bi_sc=2^(cal%(&hd9) and 15)
  Bcp=cal%(&hd5):if Bcp>127 then inc Bcp,-256'2th complement
  Bcp=Bcp/Bi_sc
  TGC=cal%(&hd8):if TGC>127 then inc TGC,-256'2th complement
  dAi_sc=2^(cal%(&hd9)\16)
  Vth=cal%(&hda)+256*cal%(&hdb):if Vth>32767 then inc Vth,-65536'2th complement
  Kt1=cal%(&hdc)+256*cal%(&hdd):if Kt1>32767 then inc Kt1,-65536'2th complement
  Kt2=cal%(&hde)+256*cal%(&hdf):if Kt2>32767 then inc Kt2,-65536'2th complement
  AL0=cal%(&he0)+256*cal%(&he1)
  AL0_sc=2^cal%(&he2)
  dAL_sc=2^cal%(&he3)
  ALcp=cal%(&hd6)+256*cal%(&hd7):if ALcp>32767 then inc ALcp,-65536'2th complement
  Eps=(cal%(&he4)+cal%(&he5)*256)/32768
  
end sub
  
  'initialize oscillator to correct frequency
sub write_osc
  
  local Trim_AA%=256+osctrim-&haa, cmmd%=4
  i2c write SA , 0 , 5 , cmmd% , Trim_AA% , osctrim , &h56 , 0
  
  'verify written
  if debug then
    cmmd%=2
    i2c write SA , 1 , 4 , cmmd% , &h93 , 0 , 1   'read 1 register from 0x93
    I2C read SA , 0 , 2 , da%()
    print "OSC_CAL = &h";hex$(da%(0)+256*da%(1))
  end if
  
end sub
  
  'initialize chip configuration
sub write_conf
  
  local cmmd%=&h3, ConfL_55=256 + Conf_L - &h55, ConfH_55=256 + Conf_H - &h55
  'print hex$(ConfL_55),hex$(ConfH_55)
  i2c write SA , 0 , 5 , cmmd% , ConfL_55 , Conf_L , ConfH_55 , Conf_H
  
  'verify written
  if debug then
    cmmd%=2
    i2c write SA , 1 , 4 , cmmd% , &h92 , 0 , 1   'read 1 register from 0x92
    I2C read SA , 0 , 2 , da%()
    print "Conf reg = &h";hex$(da%(0)+256*da%(1))
  end if
  
end sub
  
  'checks if there has been a brown_out or powere on reset
function POR% as integer
  
  'returns 1 when all okay, and 0 when the system needs re-init
  cmmd%=2
  i2c write SA , 1 , 4 , cmmd% , &h92 , 0 , 1   'read 1 register from 0x92
  I2C read SA , 0 , 2 , da%()
  if da%(1) and &b100 then POR%=1 else POR%=0
  
end function
  
  
  'reads the compensation pixel value from 0x41
sub compix_read
  cmmd%=2
  i2c write SA , 1 , 4 , cmmd% , &h41 , 0 , 1   'read 1 register from 0x41
  I2C read SA , 0 , 2 , da%()
  Vcp = da%(0)+256*da%(1):if Vcp>32767 then inc Vcp,-65536
  if debug then print "Comp_Pix = ";Vcp
  
end sub
  
  'reads the ambient temperature value from 0x40
sub ptat_read
  cmmd%=2
  i2c write SA , 1 , 4 , cmmd% , &h40 , 0 , 1   'read 1 register from 0x40
  I2C read SA , 0 , 2 , da%()
  ptat_val = da%(0)+256*da%(1):if ptat_val>32767 then inc ptat_val,-65536
  if debug then print "PTAT_val = ";ptat_val
  
end sub
  
  'calculate ambient temperature from EEPROM data and package temp read from the device
sub calc_ta
  
  'local k_t1!,k_t2!
  k_t1!=Kt1/(2^(KTscale\16))
  k_t2!=Kt2/(2^((KTscale and 15)+10))
  Ta=((sqr(k_t1!^2 - 4*k_t2!*(Vth-ptat_val))-k_t1!)/(2*k_t2!))+25
  if debug then print "Kt1=";k_t1!,"Kt2=";k_t2!;"Ta=";Ta
  
end sub
